"""
Application global configuration.
"""
from dataclasses import dataclass

from .interfaces import IStdLibFilter


class AppConfigurationAlreadySetException(Exception):
    """
    Exception when you try to set app configuration twice.
    """


@dataclass(frozen=True)
class AppConfiguration:
    """
    Configuration define at application level (in general in main).
    """

    std_lib_filter: IStdLibFilter


class AppConfigurationSingleton:
    """
    App configuration singleton
    """

    _instance: AppConfiguration
    _already_set: bool = False

    @classmethod
    def define_app_configuration(cls, configuration: AppConfiguration) -> None:
        """
        Define the application configuration to use during this run.
        Can only be called once.
        """
        if cls._already_set:
            raise AppConfigurationAlreadySetException(
                "app configuration can be set once"
            )
        cls._instance = configuration
        cls._already_set = True

    @classmethod
    def get_instance(cls) -> AppConfiguration:
        """
        Get current app configuration instance.
        raise if not define.
        """
        return cls._instance



"""
Test graph use case
"""
from typing import Iterator
from unittest.mock import Mock, patch

from dep_check.infra.io import Graph, GraphDrawer
from dep_check.infra.python_parser import PythonParser
from dep_check.models import Dependency, Module, SourceFile
from dep_check.use_cases.draw_graph import DrawGraphUC, _fold_dep

from .fakefile import GLOBAL_DEPENDENCIES, SIMPLE_FILE

PARSER = PythonParser()


def test_empty_source_files() -> None:
    """
    Test result with no source files given.
    """
    # Given
    source_files: Iterator[SourceFile] = iter([])
    drawer = Mock()
    use_case = DrawGraphUC(drawer, PARSER, source_files)

    # When
    use_case.run()

    # Then
    drawer.write.assert_called_with({})


def test_nominal(source_files) -> None:
    """
    Test result with a set source files.
    """
    # Given
    drawer = Mock()
    use_case = DrawGraphUC(drawer, PARSER, source_files)

    # When
    use_case.run()

    # Then
    drawer.write.assert_called()  # type: ignore
    global_dep = drawer.write.call_args[0][0]

    assert global_dep == GLOBAL_DEPENDENCIES


def test_dot() -> None:
    """
    Test that the dot file is well generated
    """
    # Given
    source_files: Iterator[SourceFile] = iter([SIMPLE_FILE])
    drawer = GraphDrawer(Graph("tests/graph.svg"))
    use_case = DrawGraphUC(drawer, PARSER, source_files)

    # When
    use_case.run()

    # Then
    with open("/tmp/graph.dot", encoding="utf-8") as dot:
        lines = sorted(dot.readlines())

    assert lines == sorted(
        [
            "digraph G {\n",
            "splines=true;\n",
            f"node[shape=box fontname=Arial style=filled fillcolor={drawer.graph.node_color}];\n",
            f"bgcolor={drawer.graph.background_color}\n",
            "\n",
            '"simple_module" -> "module"\n',
            '"simple_module" -> "module.inside.module"\n',
            '"simple_module" -> "amodule"\n',
            "}\n",
        ]
    )


@patch.object(GraphDrawer, "_write_svg")
def test_not_svg_with_dot(mock_method) -> None:
    """
    Test that no svg file is created when .dot in argument
    """
    # Given
    source_files: Iterator[SourceFile] = iter([SIMPLE_FILE])
    drawer = GraphDrawer(Graph("tests/graph.dot"))
    use_case = DrawGraphUC(drawer, PARSER, source_files)

    # When
    use_case.run()

    # Then
    mock_method.assert_not_called()


def test_fold_dep_empty_dict() -> None:
    """
    Test result of _fold_dep function with an empty dictionary
    """
    # Given
    global_dep = {}
    fold_module = Module("module")

    # When
    global_dep = _fold_dep(global_dep, fold_module)

    # Then
    global_dep = {}


def test_fold_dep_empty_module() -> None:
    """
    Test result of _fold_dep function with an empty fold module
    """
    # Given
    global_dep = GLOBAL_DEPENDENCIES
    fold_module = Module("")

    # When
    global_dep = _fold_dep(global_dep, fold_module)

    # Then
    global_dep = GLOBAL_DEPENDENCIES


def test_fold_dep() -> None:
    """
    Test result of _fold_dep function with an empty fold module
    """
    # Given
    global_dep = GLOBAL_DEPENDENCIES
    fold_module = Module("amodule")

    # When
    global_dep = _fold_dep(global_dep, fold_module)

    # Then
    assert global_dep == {
        "simple_module": set(
            (
                Dependency(Module("module")),
                Dependency(Module("module.inside.module")),
                Dependency(Module("amodule")),
            )
        ),
        "amodule": set(
            (
                Dependency(Module("module")),
                Dependency(Module("module.inside.module")),
                Dependency(Module("amodule")),
            )
        ),
    }


def test_fold_module(source_files) -> None:
    """
    Test result with a set source files and a module to fold.
    """
    # Given
    conf_graph = {"fold_modules": ["amodule"]}
    drawer = Mock()
    use_case = DrawGraphUC(drawer, PARSER, source_files, conf_graph)

    # When
    use_case.run()

    # Then
    drawer.write.assert_called()  # type: ignore
    global_dep = drawer.write.call_args[0][0]

    assert global_dep == {
        "simple_module": set(
            (
                Dependency(Module("module")),
                Dependency(Module("module.inside.module")),
                Dependency(Module("amodule")),
            )
        ),
        "amodule": set(
            (Dependency(Module("module")), Dependency(Module("module.inside.module")))
        ),
    }


@patch.object(DrawGraphUC, "_hide")
def test_hide_empty_config(mock_method, source_files) -> None:

    # Given
    drawer = Mock()
    config = {"fold_modules": ["module"]}
    use_case = DrawGraphUC(drawer, PARSER, source_files, config)

    # When
    use_case.run()

    # Then
    mock_method.assert_called_with(GLOBAL_DEPENDENCIES)


@patch.object(DrawGraphUC, "_hide")
def test_hide_empty_dict(mock_method) -> None:

    # Given
    source_files: Iterator[SourceFile] = iter([])
    drawer = Mock()
    config = {"hide_modules": ["toto", "tata"]}
    use_case = DrawGraphUC(drawer, PARSER, source_files, config)

    # When
    use_case.run()

    # Then
    mock_method.assert_called_with({})


def test_hide_nominal(source_files) -> None:
    # Given
    drawer = Mock()
    config = {"hide_modules": ["amodule"]}
    use_case = DrawGraphUC(drawer, PARSER, source_files, config)

    # When
    use_case.run()

    # Then
    drawer.write.assert_called()  # type: ignore
    global_dep = drawer.write.call_args[0][0]

    assert global_dep == {
        "simple_module": set(
            (Dependency(Module("module")), Dependency(Module("module.inside.module")))
        )
    }


def test_pop_empty_module_from_dependencies(source_files) -> None:
    # Given
    drawer = Mock()
    config = {"hide_modules": ["module"]}
    use_case = DrawGraphUC(drawer, PARSER, source_files, config)

    # When
    use_case.run()

    # Then
    drawer.write.assert_called_with(
        {
            "simple_module": set((Dependency(Module("amodule")),)),
            "amodule.local_module": set(
                (Dependency(Module("amodule")), Dependency(Module("amodule.inside")))
            ),
        }
    )



from dep_check.models import Dependency, Module, SourceCode, SourceFile

SIMPLE_FILE = SourceFile(
    module=Module("simple_module"),
    code=SourceCode(
        """
import module
import module.inside.module
from amodule import aclass
"""
    ),
)
FILE_WITH_LOCAL_IMPORT = SourceFile(
    module=Module("amodule.local_module"),
    code=SourceCode(
        """
import module
import module.inside.module
from . import aclass
from .inside import aclass
"""
    ),
)
FILE_WITH_STD_IMPORT = SourceFile(
    module=Module("amodule.std_module"),
    code=SourceCode(
        """
import module
import module.inside.module
import itertools
from abc import ABC
"""
    ),
)

GLOBAL_DEPENDENCIES = {
    "simple_module": set(
        (
            Dependency(Module("module")),
            Dependency(Module("module.inside.module")),
            Dependency(Module("amodule")),
        )
    ),
    "amodule.local_module": set(
        (
            Dependency(Module("module")),
            Dependency(Module("module.inside.module")),
            Dependency(Module("amodule")),
            Dependency(Module("amodule.inside")),
        )
    ),
    "amodule.std_module": set(
        (Dependency(Module("module")), Dependency(Module("module.inside.module")))
    ),
}



"""
Check all given source files dependencies use case.
"""


import re
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Iterator, List, Tuple

from dep_check.checker import NotAllowedDependencyException, check_dependency
from dep_check.dependency_finder import IParser, get_import_from_dependencies
from dep_check.models import Module, ModuleWildcard, Rules, SourceFile

from .app_configuration import AppConfigurationSingleton
from .interfaces import Configuration


class ForbiddenError(Exception):
    pass


class ForbiddenUnusedRuleError(ForbiddenError):
    pass


class ForbiddenDepencyError(ForbiddenError):
    pass


@dataclass(frozen=True)
class DependencyError:
    """
    Dataclass representing a dependency error.
    """

    module: Module
    dependency: Module
    rules: Tuple[ModuleWildcard, ...]


class IReportPrinter(ABC):
    """
    Errors printer interface.
    """

    @abstractmethod
    def _error(self, dep_errors: List[DependencyError]) -> None:
        """
        Display errors.
        """

    @abstractmethod
    def _warning(self, unused_rules: Rules) -> None:
        """
        Display warnings.
        """

    @abstractmethod
    def print_report(
        self, errors: List[DependencyError], unused_rules: Rules, nb_files: int
    ) -> None:
        """
        Print report
        """


class CheckDependenciesUC:
    """
    Dependency check use case.

    In this use case, we ensure that all given source files respect
    all rules that matching their module name.
    """

    def __init__(
        self,
        configuration: Configuration,
        report_printer: IReportPrinter,
        parser: IParser,
        source_files: Iterator[SourceFile],
    ):
        app_configuration = AppConfigurationSingleton.get_instance()
        self.std_lib_filter = app_configuration.std_lib_filter
        self.configuration = configuration
        self.report_printer = report_printer
        self.parser = parser
        self.source_files = source_files
        self.used_rules: Rules = set()
        self.all_rules: Rules = set()

    def _get_rules(self, module: Module) -> Rules:
        """
        Return rules in configuration that match a given module.
        """
        matching_rules: Rules = set()
        for module_wildcard, rules in self.configuration.dependency_rules.items():
            if re.match(
                f"{self.parser.wildcard_to_regex(ModuleWildcard(module_wildcard))}$",
                module,
            ):
                matching_rules.update(
                    (ModuleWildcard(module_wildcard), r) for r in rules
                )

        return matching_rules

    def _iter_error(self, source_file: SourceFile) -> Iterator[DependencyError]:
        rules = self._get_rules(source_file.module)
        self.all_rules.update(set(rules))
        dependencies = get_import_from_dependencies(source_file, self.parser)
        dependencies = self.std_lib_filter.filter(dependencies)
        for dependency in dependencies:
            try:
                self.used_rules |= check_dependency(self.parser, dependency, rules)
            except NotAllowedDependencyException as error:
                yield DependencyError(
                    source_file.module,
                    error.dependency,
                    tuple(sorted(error.authorized_modules)),
                )

    def run(self) -> None:
        errors = []
        nb_files = 0

        for source_file in self.source_files:
            nb_files += 1
            for error in self._iter_error(source_file):
                errors.append(error)

        unused = self.all_rules.difference(self.used_rules)
        self.report_printer.print_report(errors, unused, nb_files)

        if self.configuration.error_on_unused and unused:
            raise ForbiddenUnusedRuleError

        if errors:
            raise ForbiddenDepencyError



"""
Test check use case.
"""
from typing import List

from pytest import raises

from dep_check.checker import NotAllowedDependencyException, check_dependency
from dep_check.infra.python_parser import PythonParser
from dep_check.models import Dependency, Module, ModuleWildcard, Rules

PARSER = PythonParser()


def test_empty() -> None:
    """
    Test empty rules case.
    """
    # Given
    dependency = Dependency(Module("toto"))
    authorized_modules: List[ModuleWildcard] = []

    # When
    with raises(NotAllowedDependencyException) as error:
        check_dependency(PARSER, dependency, authorized_modules)

    # Then
    assert error
    assert error.value.dependency == dependency.main_import
    assert error.value.authorized_modules == authorized_modules


def test_passing_case() -> None:
    """
    Test a passing case.
    """
    # Given
    dependency = Dependency(Module("toto"))
    rules: Rules = [
        (ModuleWildcard("toto"), ModuleWildcard("to*")),
        (ModuleWildcard("toto"), ModuleWildcard("titi.tata")),
    ]

    # When
    error = None
    try:
        check_dependency(PARSER, dependency, rules)
    except NotAllowedDependencyException as exception:
        error = exception

    # Then
    assert not error


def test_not_passing_case() -> None:
    """
    Test a not passing case.
    """
    # Given
    dependency = Dependency(Module("toto.tata"))
    rules: Rules = [
        (ModuleWildcard("toto.*"), ModuleWildcard("toto")),
        (ModuleWildcard("toto.*"), ModuleWildcard("te.*")),
        (ModuleWildcard("toto.*"), ModuleWildcard("titi\\.tata")),
    ]

    # When
    with raises(NotAllowedDependencyException) as error:
        check_dependency(PARSER, dependency, rules)

    # Then
    assert error
    assert error.value.dependency == dependency.main_import
    assert error.value.authorized_modules == [r for _, r in rules]



"""
Test build configuration use case.
"""

from typing import Iterator
from unittest.mock import Mock

from dep_check.infra.python_parser import PythonParser
from dep_check.models import ModuleWildcard, SourceFile
from dep_check.use_cases.build import BuildConfigurationUC
from dep_check.use_cases.interfaces import Configuration

PARSER = PythonParser()


def test_empty() -> None:
    """
    Test result with no source files given.
    """
    # Given
    source_files: Iterator[SourceFile] = iter([])
    dependencies_writer = Mock()
    use_case = BuildConfigurationUC(dependencies_writer, PARSER, source_files, "python")

    # When
    use_case.run()

    # Then
    dependencies_writer.write.assert_called_with(Configuration())


def test_nominal(source_files) -> None:
    """
    Test result with a set source files.
    """
    # Given
    dependencies_writer = Mock()
    use_case = BuildConfigurationUC(dependencies_writer, PARSER, source_files, "python")

    # When
    use_case.run()

    # Then
    dependencies_writer.write.assert_called()  # type: ignore
    configuration = dependencies_writer.write.call_args[0][0]
    assert not configuration.local_init
    dependency_rules = {
        module_regex: set(rules)
        for module_regex, rules in configuration.dependency_rules.items()
    }
    assert dependency_rules == {
        "simple_module": set(
            (
                ModuleWildcard("module"),
                ModuleWildcard("module.inside.module"),
                ModuleWildcard("amodule"),
            )
        ),
        "amodule.local_module": set(
            (
                ModuleWildcard("module"),
                ModuleWildcard("module.inside.module"),
                ModuleWildcard("amodule"),
                ModuleWildcard("amodule.inside"),
            )
        ),
        "amodule.std_module": set(
            (ModuleWildcard("module"), ModuleWildcard("module.inside.module"))
        ),
    }



"""
Implementation of configuration reader and writer
"""
import os
from contextlib import contextmanager
from pathlib import Path
from typing import Iterator

from dep_check.models import Module, SourceCode, SourceFile


def _get_module_from_file_path(path: Path, separator: str) -> Module:
    """
    Transform a filename into a corresponding python module.
    """
    path_without_extention = path.parents[0] / path.stem
    return Module(separator.join(path_without_extention.parts))


@contextmanager
def _change_dir(directory: str) -> Iterator[None]:
    """
    Locally change current working directory
    """
    saved_dir = os.getcwd()
    try:
        os.chdir(directory)
        yield
    finally:
        os.chdir(saved_dir)


def _get_python_project_root(root_dir: str) -> str:
    """
    Returns the project root to make sure every module name begins with it
    """
    project_root = ""

    if "__init__.py" in os.listdir(Path(root_dir)):
        project_root = Path(root_dir).name + "."

    for directory in Path(root_dir).parents:
        if "__init__.py" in os.listdir(directory):
            project_root = f"{directory.name}.{project_root}"
        else:
            break
    return project_root


def source_file_iterator(root_dir: str, file_extension: str) -> Iterator[SourceFile]:
    """
    Iterator of all python source files in a directory.
    """
    if file_extension == "py":
        project_root = _get_python_project_root(root_dir)
        separator = "."
    elif file_extension == "go":
        project_root = ""
        separator = "/"

    with _change_dir(root_dir):
        for file_path in Path(".").rglob(f"*.{file_extension}"):
            with open(str(file_path), "r", encoding="utf-8") as stream:
                content = stream.read()
            yield SourceFile(
                Module(project_root + _get_module_from_file_path(file_path, separator)),
                SourceCode(content),
            )



from abc import ABC, abstractmethod

from dep_check.models import Dependencies, ModuleWildcard, SourceFile


class IParser(ABC):
    """
    Interface for the code parser
    """

    @abstractmethod
    def wildcard_to_regex(self, module: ModuleWildcard) -> str:
        """
        Return a regex expression for the Module from wildcard
        """

    @abstractmethod
    def find_dependencies(self, source_file: SourceFile) -> Dependencies:
        """
        Find the source files' dependencies
        """

    @abstractmethod
    def find_import_from_dependencies(self, source_file: SourceFile) -> Dependencies:
        """
        Find the source files' from... import" dependencies
        """


def get_dependencies(source_file: SourceFile, parser: IParser) -> Dependencies:
    return parser.find_dependencies(source_file)


def get_import_from_dependencies(
    source_file: SourceFile, parser: IParser
) -> Dependencies:
    return parser.find_import_from_dependencies(source_file)



"""
Draws a graph of the dependencies between modules
"""
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Dict, Iterator, Optional

from dep_check.dependency_finder import IParser, get_dependencies
from dep_check.models import (
    Dependencies,
    Dependency,
    GlobalDependencies,
    Module,
    SourceFile,
)

from .app_configuration import AppConfigurationSingleton


class IGraphDrawer(ABC):
    """
    Interface for writing a dependency graph.
    """

    @abstractmethod
    def write(self, global_dep: GlobalDependencies) -> bool:
        """
        Writes a dot file corresponding to the dependency graph.
        """


def _fold_dep(
    global_dep: GlobalDependencies, fold_module: Module
) -> GlobalDependencies:

    fold_global_dep: GlobalDependencies = defaultdict(set)
    for module, deps in global_dep.items():
        new_deps: Dependencies = set()
        if module.startswith(tuple((f"{fold_module}.", f"{fold_module}/"))):
            module = fold_module
        fold_dep = set(
            Dependency(fold_module) if dep.main_import.startswith(fold_module) else dep
            for dep in deps
        )
        new_deps |= fold_dep
        fold_global_dep[module] |= new_deps

    return fold_global_dep


class DrawGraphUC:
    def __init__(
        self,
        drawer: IGraphDrawer,
        parser: IParser,
        source_files: Iterator[SourceFile],
        config: Optional[Dict] = None,
    ):
        app_configuration = AppConfigurationSingleton.get_instance()
        self.std_lib_filter = app_configuration.std_lib_filter
        self.source_files = source_files
        self.drawer = drawer
        self.parser = parser
        self.config = config or {}

    def _hide(self, global_dep: GlobalDependencies) -> GlobalDependencies:
        hide_modules = tuple(self.config.get("hide_modules", ()))
        if not hide_modules:
            return global_dep

        filtered_global_dep = {}
        for module, dependencies in global_dep.items():
            if not module.startswith(hide_modules):

                filtered_global_dep[module] = set(
                    dependency
                    for dependency in dependencies
                    if not dependency.main_import.startswith(hide_modules)
                )

        return filtered_global_dep

    def run(self) -> None:
        global_dependencies: GlobalDependencies = {}
        for source_file in self.source_files:
            module = Module(source_file.module.replace(".__init__", ""))
            dependencies = get_dependencies(source_file, self.parser)
            dependencies = self.std_lib_filter.filter(dependencies)
            global_dependencies[module] = dependencies

        global_dependencies = self._hide(global_dependencies)

        for fold_module in self.config.get("fold_modules", []):
            global_dependencies = _fold_dep(global_dependencies, fold_module)

        # To avoid a module to point itself, and make the graph more readable
        for module, deps in list(global_dependencies.items()):
            deps -= {dep for dep in deps if dep.main_import == module}

            if not global_dependencies[module]:
                global_dependencies.pop(module, None)

        self.drawer.write(global_dependencies)



"""
Test check use case.
"""

from unittest.mock import Mock

import pytest

from dep_check.infra.python_parser import PythonParser
from dep_check.models import Module, ModuleWildcard, SourceCode, SourceFile
from dep_check.use_cases.check import (
    CheckDependenciesUC,
    DependencyError,
    ForbiddenDepencyError,
    ForbiddenUnusedRuleError,
)
from dep_check.use_cases.interfaces import Configuration

from .fakefile import FILE_WITH_LOCAL_IMPORT, FILE_WITH_STD_IMPORT, SIMPLE_FILE

PARSER = PythonParser()


def test_empty_rules(source_files) -> None:
    """
    Test result with no rule given.
    """
    # Given
    configuration = Configuration()
    report_printer = Mock()
    use_case = CheckDependenciesUC(configuration, report_printer, PARSER, source_files)

    # When
    with pytest.raises(ForbiddenDepencyError):
        use_case.run()

    # Then
    assert set(report_printer.print_report.call_args[0][0]) == set(
        (
            DependencyError(SIMPLE_FILE.module, Module("module"), tuple()),
            DependencyError(
                SIMPLE_FILE.module, Module("module.inside.module"), tuple()
            ),
            DependencyError(SIMPLE_FILE.module, Module("amodule.aclass"), tuple()),
            DependencyError(FILE_WITH_LOCAL_IMPORT.module, Module("module"), tuple()),
            DependencyError(
                FILE_WITH_LOCAL_IMPORT.module, Module("module.inside.module"), tuple()
            ),
            DependencyError(
                FILE_WITH_LOCAL_IMPORT.module, Module("amodule.aclass"), tuple()
            ),
            DependencyError(
                FILE_WITH_LOCAL_IMPORT.module, Module("amodule.inside.aclass"), tuple()
            ),
            DependencyError(FILE_WITH_STD_IMPORT.module, Module("module"), tuple()),
            DependencyError(
                FILE_WITH_STD_IMPORT.module, Module("module.inside.module"), tuple()
            ),
        )
    )


def test_passing_rules(source_files) -> None:
    """
    Test result with a set rules that accept files.
    """
    # Given
    configuration = Configuration(
        dependency_rules={
            SIMPLE_FILE.module: [ModuleWildcard("module%"), ModuleWildcard("amodule")],
            "amodule.*": [
                ModuleWildcard("module"),
                ModuleWildcard("module.inside.*"),
                ModuleWildcard("amodule%"),
                ModuleWildcard("unused%"),
            ],
        }
    )
    report_printer = Mock()
    use_case = CheckDependenciesUC(configuration, report_printer, PARSER, source_files)

    # When
    use_case.run()

    # Then
    report_printer.print_report.assert_called_with(
        [], set(((ModuleWildcard("amodule.*"), ModuleWildcard("unused%")),)), 3
    )


def test_error_on_unused(source_files) -> None:
    """
    Test result with a set rules that accept files.
    """
    # Given
    configuration = Configuration(
        dependency_rules={
            SIMPLE_FILE.module: [ModuleWildcard("module%"), ModuleWildcard("amodule")],
            "amodule.*": [
                ModuleWildcard("module"),
                ModuleWildcard("module.inside.*"),
                ModuleWildcard("amodule%"),
                ModuleWildcard("unused%"),
            ],
        },
        error_on_unused=True,
    )
    report_printer = Mock()
    use_case = CheckDependenciesUC(configuration, report_printer, PARSER, source_files)

    # When
    with pytest.raises(ForbiddenUnusedRuleError):
        use_case.run()

    # Then
    report_printer.print_report.assert_called_with(
        [], set(((ModuleWildcard("amodule.*"), ModuleWildcard("unused%")),)), 3
    )


def test_not_passing_rules(source_files) -> None:
    """
    Test result with a set rules that not accept files.
    """
    # Given
    dep_rules = {
        "simple_module": [ModuleWildcard("module.*"), ModuleWildcard("amodule")],
        "amodule.local_module": [
            ModuleWildcard("module"),
            ModuleWildcard("module.inside.*"),
            ModuleWildcard("amod"),
        ],
        "amodule.std_module": [ModuleWildcard("mod")],
    }

    configuration = Configuration(dependency_rules=dep_rules)
    report_printer = Mock()
    use_case = CheckDependenciesUC(configuration, report_printer, PARSER, source_files)

    # When
    with pytest.raises(ForbiddenDepencyError):
        use_case.run()

    # Then
    simple = SIMPLE_FILE.module
    local = FILE_WITH_LOCAL_IMPORT.module
    std = FILE_WITH_STD_IMPORT.module

    assert set(report_printer.print_report.call_args[0][0]) == set(
        (
            DependencyError(
                simple, Module("module"), tuple(sorted(dep_rules["simple_module"]))
            ),
            DependencyError(
                local,
                Module("amodule.aclass"),
                tuple(sorted(dep_rules["amodule.local_module"])),
            ),
            DependencyError(
                local,
                Module("amodule.inside.aclass"),
                tuple(sorted(dep_rules["amodule.local_module"])),
            ),
            DependencyError(
                std, Module("module"), tuple(sorted(dep_rules["amodule.std_module"]))
            ),
            DependencyError(
                std,
                Module("module.inside.module"),
                tuple(sorted(dep_rules["amodule.std_module"])),
            ),
        )
    )


def test_passing_rules_with_import_from() -> None:
    """
    Test result with a set rules that accept files.
    """
    # Given
    configuration = Configuration(
        dependency_rules={
            "module": [ModuleWildcard("module%"), ModuleWildcard("amodule.submodule")]
        }
    )
    source_file = SourceFile("module", SourceCode("from amodule import submodule"))
    report_printer = Mock()
    use_case = CheckDependenciesUC(
        configuration, report_printer, PARSER, iter([source_file])
    )

    # When
    use_case.run()

    # Then
    assert report_printer.print_report.call_args[0][0] == []


def test_not_passing_rules_with_import_from() -> None:
    """
    Test result with a set rules that accept files.
    """
    # Given
    dep_rules = {
        "module": [ModuleWildcard("module%"), ModuleWildcard("amodule.submodule")]
    }
    configuration = Configuration(dep_rules)
    source_file = SourceFile(
        "module", SourceCode("from amodule import othermodule, submodule\n")
    )
    report_printer = Mock()
    use_case = CheckDependenciesUC(
        configuration, report_printer, PARSER, iter([source_file])
    )

    # When
    with pytest.raises(ForbiddenDepencyError):
        use_case.run()

    # Then
    assert set(report_printer.print_report.call_args[0][0]) == set(
        (
            DependencyError(
                "module",
                Module("amodule.othermodule"),
                tuple(sorted(dep_rules["module"])),
            ),
        )
    )



"""
Check that dependencies follow a set of rules.
"""
import re
from typing import List, Optional

from dep_check.dependency_finder import IParser
from dep_check.models import Dependency, Module, ModuleWildcard, Rule, Rules


class NotAllowedDependencyException(Exception):
    """
    Exception that represents an unauthorized import in a module.
    """

    def __init__(
        self, dependency: Module, authorized_modules: List[ModuleWildcard]
    ) -> None:
        super().__init__(
            f"try to import {dependency} but only allowed to import {authorized_modules}"
        )
        self.dependency = dependency
        self.authorized_modules = authorized_modules


def check_dependency(parser: IParser, dependency: Dependency, rules: Rules) -> Rules:
    """
    Check that dependencies match a given set of rules.
    """
    used_rule: Optional[Rule] = None
    for module, rule in rules:
        if re.match(f"{parser.wildcard_to_regex(rule)}$", dependency.main_import):
            used_rule = (module, rule)
            return set((used_rule,))
    if not dependency.sub_imports:
        raise NotAllowedDependencyException(
            dependency.main_import, [r for _, r in rules]
        )

    return check_import_from_dependency(parser, dependency, rules)


def check_import_from_dependency(
    parser: IParser, dependency: Dependency, rules: Rules
) -> Rules:
    used_rules: Rules = set()
    for import_module in dependency.sub_imports:
        used_rule = None
        for module, rule in rules:
            if re.match(
                f"{parser.wildcard_to_regex(rule)}$",
                f"{dependency.main_import}.{import_module}",
            ):
                used_rule = (module, rule)
                used_rules.add(used_rule)
        if not used_rule:
            raise NotAllowedDependencyException(
                Module(f"{dependency.main_import}.{import_module}"),
                [r for _, r in rules],
            )
    return used_rules



"""
Test functions in models module.
"""

from pytest import raises

from dep_check.models import Module, get_parent, iter_all_modules

from .fakefile import GLOBAL_DEPENDENCIES


class TestGetParent:
    """
    Test get_parent function.
    """

    @staticmethod
    def test_empty() -> None:
        """
        Test empty case.
        """
        # Given
        module = Module("")

        # When
        parent = get_parent(module)

        # Then
        assert parent == ""

    @staticmethod
    def test_simple_module() -> None:
        """
        Test simple case.
        """
        # Given
        module = Module("toto")

        # When
        parent = get_parent(module)

        # Then
        assert parent == Module("")

    @staticmethod
    def test_nested_module() -> None:
        """
        Test nested case.
        """
        # Given
        module = Module("toto.tata")

        # When
        parent = get_parent(module)

        # Then
        assert parent == Module("toto")

    @staticmethod
    def test_long_nested_module() -> None:
        """
        Test long nested case.
        """
        # Given
        module = Module("toto.titi.tete.tata")

        # When
        parent = get_parent(module)

        # Then
        assert parent == Module("toto.titi.tete")


class TestIterAllModules:
    """
    Test iter_all_modules function
    """

    @staticmethod
    def test_empty() -> None:
        """
        Test empty case.
        """
        # Given
        global_deps = {}

        # When
        result_iter = iter_all_modules(global_deps)

        # Then
        with raises(StopIteration):
            next(result_iter)

    @staticmethod
    def test_nominal() -> None:
        """
        Test nominal case.
        """
        # Given
        global_deps = GLOBAL_DEPENDENCIES

        # When
        result_iter = iter_all_modules(global_deps)

        # Then
        all_modules = []
        for module in result_iter:
            all_modules.append(module)

        assert sorted(all_modules) == sorted(
            [
                "simple_module",
                "module",
                "module.inside.module",
                "amodule",
                "amodule.local_module",
                "amodule.inside",
                "amodule.std_module",
            ]
        )
